/*
 * iqs253.c - Android SAR Sensor Driver for IQS253
 *
 * Copyright (C) 2013 Azoteq (Pty) Ltd
 * Author: Alwino van der Merwe <alwino.vandermerwe@azoteq.com>
 *
 * Based on mcs5000_ts.c, azoteqiqs440_ts.c, iqs263.c
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *	Azoteq (Pty) Ltd does not take responsibility for the use of this driver
 *
 *	This driver is an example driver. It will need to be ported to
 *	the specific platform and for the specific case in which it is used.
 */
#include "iqs253.h"
#include "IQS253_Init.h"

#include <linux/module.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/irq.h>
#include <linux/gpio.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>

/*	Get time for the IQS253 Timeout	*/
#include <linux/time.h>

/****************************
 *	Pins used in this setup	*
 ****************************/

/*	RDY Pin used on this specific setup	- needs to be changed
 *	for each case (hardware dependent)
 */
/*	GPIO0_27 - Bank 0	*/
#define RDY_LINE_PIN			27

/*	VDDHI used on this specific setup	- needs to be changed
 *	for each case (hardware dependent).
 *	The SAR sensor can also be powered from the voltage
 *	rail of the system, but this allows more freedom.
 */
/*	GPIO0_7 - Bank 0 Header P9-42 on BBB	*/
#define PWR_LINE_PIN			7

/*	Numbet of Bytes to read from IQS253 continuously	*/
#define IQS253_BLOCK_SIZE		3
#define IQS253_SETUP_REG_SIZE	26

/* Boolean value used for the initial setup */
bool doInitialSetup;
bool initialATI;
/*	Indicate the a Reseed of the IQS253 will follow	*/
bool reseed;

u8 events;

/*	Boolean to keep track of Chip Reset Events	*/
bool showReset;
/*	Boolean to keep track of Prox   Events	*/
bool prox;
/*	Boolean to keep track of Touch Events	*/
bool touch;

/*	Counter to keep track of setup state	*/
u8 setupCounter;

/*	Global variable to keep the currentState in	*/
u8 currentState;	/*	Always start at state 0	*/

/*	Boolean to indicate event mode active or not	*/
bool eventMode;		/*	not activated by default	*/

/*	Each client has this additional data	*/
struct iqs253_sar_data {
	struct i2c_client *client;
	struct input_dev *input_dev;
	const struct iqs253_sar_platform_data *platform_data;
};

/*	Struct to keep the timer SAR Timer information in	*/
static struct timer_list stuck_timer;

/*	Struct to keep a stuck timer in	*/
static struct timer_list stuck;

/********************************************************
 *		The IQS253 specific setup is done here			*
 *		This setup will change for each application		*
 ********************************************************/

/*	Command for reseeding the Azoteq IQS253	*/
static void iqs253_reseed(struct iqs253_sar_data *data)
{
	i2c_smbus_write_byte_data(data->client, PROX_SETTINGS0, RESEED);
}

/*	Command for re-doing ATI on the Azoteq IQS253	*/
static void iqs253_reati(struct iqs253_sar_data *data)
{
	/*	Redo ATI	*/
	i2c_smbus_write_byte_data(data->client, PROX_SETTINGS0, REDO_ATI);
}


/************************************************************************
 *							IQS253 Initial Setup						*
 ************************************************************************/
/* Setup the Active Channels for Azoteq IQS253 */
static void iqs253_init_setup(struct iqs253_sar_data *data, u8 numOfRegs)
{
	/**
	 *	Setup IQS253 completely - start at ATI register and
	 *	run through all regsiters
	 */
	i2c_smbus_write_i2c_block_data(data->client, ATI_TARGET, numOfRegs,
			iqs253_default_regs);
}

/************************************************************************
 *						State Machine Helper Functions					*
 ************************************************************************/

/**
 *	Check for Events on the IQS253. ATI, Prox and Touch event. Also check
 *	the Show Reset Flag
 *	The function returns no value - instead Global flags are set to
 *	indicate the events that ocurred
 */
static void readEvents(struct iqs253_sar_data *data)
{
	u8 data_buffer[3];

	/*
	 *	Read the Bytes for Prox and Touch events -
	 *	save this state
	 */
	i2c_smbus_read_i2c_block_data(data->client, DEFAULT_COMMS_POINTER_VAL,
		IQS253_BLOCK_SIZE, data_buffer);

	/*	We are interested in byte 0, byte 1 and byte 2	*/
	/*	These bytes will give us all of the info we need	*/
	if (data_buffer[0]&SHOW_RESET)
		showReset = true;
	else
		showReset = false;

	if (data_buffer[1]&CH_0_EVENT)
		prox = true;
	else
		prox = false;

	if (data_buffer[2]&CH_0_EVENT)
		touch = true;
	else
		touch = false;
}

/*************************************************************************/

/*	Platform data for the Azoteq IQS253 SAR driver	*/
struct iqs253_sar_platform_data {
	void (*cfg_pin)(void);
};

/**
 *	Because the IQS253 operates in event mode, implement a handshake
 *	function that will initiate comms with the IQS253 if we want to talk
 *	to it.
 */
void iqs253_event_mode_handshake(void)
{
	/********************************************************
	 *			Try and do an Event Mode Handshake			*
	 *******************************************************/
	/*
	 *	There might be another way to build in the delay.
	 *	Event mode handshake is done by manually pulling
	 *	the IQS253 RDY line low for ~10ms and then releasing
	 *	it. This will tell the IQS253 that the Master wants
	 *	to talk and it will then give a Communications window
	 *	which will call the interrupt request function.
	 *	From there the IQS253 can be read again
	 */
	/*	Pull RDY low	*/
	gpio_direction_output(RDY_LINE_PIN, 0);
	/*	Hold the RDY low for ~10ms	*/
	mdelay(HANDSHAKE_DELAY_HOLD);
	/*	Release RDY line	*/
	gpio_direction_input(RDY_LINE_PIN);
	/*	Delay before talking to IQS253	*/
	udelay(HANDSHAKE_DELAY_SET);
}

/**
 *	Function that gets called when an interrupt does not occur after a while
 */
void interrupt_timer(unsigned long data)
{
	if (currentState > 0) {
		/*	Now from here Reseed and go back to State 0	*/
		currentState = 0;
		/*	Chip will re-ATI if reseed was out of spec	*/
		reseed = true;	/*	Reseed	*/
	}

	/********************************************************
	 *			Try and do an Event Mode Handshake			*
	 *******************************************************/

	iqs253_event_mode_handshake();
}

/** Timer interrupt function	*/
void stuck_time(unsigned long data)
{
	del_timer(&stuck_timer);
	iqs253_event_mode_handshake();
}

/**
 *	Interrupt event fires on the Falling edge of
 *	the RDY signal from the Azoteq IQS253
 */
static irqreturn_t iqs253_sar_interrupt(int irq, void *dev_id)
{
	struct iqs253_sar_data *data = dev_id;

	/*	Do the initial setup for the IQS253
	 *	Because we have to wait for the RDY from the IQS253 and
	 *	we cannot change the way Android handles I2C comms
	 *	the setup will take a few comms cycles
	 */

	/*
	 *	Check for a reset - if reset was seen,
	 *	Then setup has to been done all over again
	 *
	 *	If we need to setup the IQS253, traverse through the
	 *	setup
	 */
	if (doInitialSetup) {
		switch (setupCounter) {
		/*	Setup ProxSettings	*/
		case 0:
			printk(KERN_ALERT "Initializing..");
			/*	Del timer that should have
			 *	made init window
			 */
			del_timer(&stuck);
			/*	Do event mode handshake	*/
			iqs253_init_setup(data, IQS253_SETUP_REG_SIZE);
			setupCounter++;
			break;
		case 1:
			iqs253_reati(data);
			setupCounter = 0;
			doInitialSetup = false;
			/*
			 *	Setup timer to trigger event
			 *	for ATI
			 */
			printk(KERN_ALERT "Re-ATI");
			break;
		default:
			setupCounter = 0;
			doInitialSetup = false;
			break;
		}
		goto out;
	}
	del_timer(&stuck);

	if (reseed) {
		iqs253_reseed(data);
		reseed = false;
		/*	Reset State machine	*/
		/*	Reset touch and prox status	*/
		currentState = 0;
		touch = false;
		events = 0;
		/*	Setup your timer to call my_timer_callback	*/
		/*	Ensure stuck condition does not occur	*/
		del_timer(&stuck_timer);
		del_timer(&stuck);
		/*	Setup your timer to call my_timer_callback	*/
		setup_timer(&stuck_timer, interrupt_timer, 0);
		mod_timer(&stuck_timer,
				jiffies + msecs_to_jiffies(50));
		goto out;
	}

	/*	Setup done	*/


	/**********************	State Machine	***************************/

	readEvents(data);	/*	Check Events	*/

	/*	Wait for Prox event	*/
	/*	No Activation State	*/
	if (currentState == 0) {

		/*	NOTE: Antennae or other digital/analog blocks
		 *	can be enabled here - this line only shows
		 *	where you should implement your activation
		 *	or deactivation
		 */
		/* DEBUG - LED Off	*/
		gpio_set_value(53, 0);

		if (showReset) {
			doInitialSetup = true;
			setupCounter = 0;
		}

		/*	Move to next state if Prox was seen	*/
		if (prox) {
			currentState = 1;
			del_timer(&stuck_timer);

			/*	Setup your timer to call
			 *	my_timer_callback
			 */
			setup_timer(&stuck_timer, interrupt_timer, 0);
			mod_timer(&stuck_timer,
				jiffies + msecs_to_jiffies(10000));
		} else
			currentState = 0;
	}

	/*
	 *	Activation State
	 *	Check for touch and prox - if a touch is seen,
	 *	move to the touch activation state and clear timer.
	 *	If only a prox is seen, stay in
	 *	this state and keep timer active
	 */
	if (currentState == 1) {


		/*	If activated and not a Prox - Clear	*/
		if (!prox) {

			/*	Clear Activation State	*/
			currentState = 0;
			/*	No more Prox - Clear	*/

			/*	NOTE: Antennae or other digital/analog blocks
			 *	can be enabled here - this line only shows
			 *	where you should implement your activation
			 *	or deactivation
			 */
			/* DEBUG - LED Off	*/
			gpio_set_value(53, 0);
			del_timer(&stuck_timer);
		}
		/*	Prox is still active - check touch event	*/
		else if (prox) {
			/*	NOTE: Antennae or other digital/analog blocks
			 *	can be DISABLED here - this line only shows
			 *	where you should implement your activation
			 *	or deactivation
			 */
			/*	DEBUG: Activation	*/
			gpio_set_value(53, 1);

			/* We have a Prox - Now check for touch	*/
			/*	Deep Touch	*/
			if (touch) {
				/*	Move to touch state	*/
				currentState = 2;

				del_timer(&stuck_timer);

				/*	Setup timer to call
				 *	my_timer_callback
				 */
				setup_timer(&stuck_timer, interrupt_timer, 0);
				mod_timer(&stuck_timer,
					jiffies + msecs_to_jiffies(30000));
			}
		}
	} /*	end State 1	*/

	/*
	 *	This is the Touch state - or deep activation.
	 *	We have a longer timer in this state.
	 *	When we are in this state, we want to check
	 *	the events - if no touch is seen and prox is active,
	 *	move to state 1 else move to state 0
	 *	and clear activation
	 */
	if (currentState == 2) {
		/*	Touch has disappeared	*/
		if (!touch) {
			/*
			 * Reset the timer and move to
			 * prox state if prox was seen
			 */
			/*	Prox is still active	*/
			if (prox) {
				currentState = 1;

				del_timer(&stuck_timer);

				/*	Setup your timer to call
				 *	my_timer_callback.
				 *	This timer is the SAR timeout
				 *	for a 'light' touch
				 */
				setup_timer(&stuck_timer, interrupt_timer, 0);
				mod_timer(&stuck_timer,
					jiffies + msecs_to_jiffies(10000));
			}
			/*
			 *	Now, clear prox and everything -
			 *	call reseed
			 */
			else if (!prox) {
				currentState = 0;
				reseed = true;

				/*
				 *	Now we are basically in Prox area so
				 *	release activation
				 */
				gpio_set_value(53, 0);

				/*	Setup timer to call reseed	*/
				del_timer(&stuck);
				setup_timer(&stuck, stuck_time, 0);
				mod_timer(&stuck,
					jiffies + msecs_to_jiffies(100));
			}
		}
	} /*	end State 2	*/

	/********************	State Machine End	*******************/

	/* Jump to end after each Communications window	*/
out:
	return IRQ_HANDLED;
}

/**
 *	The probe function is called when Android is looking
 *	for the IQS253 on the I2C bus (I2C-1)
 */
static int iqs253_sar_probe(struct i2c_client *client,
		const struct i2c_device_id *id)
{
	struct iqs253_sar_data *data;
	struct input_dev *input_dev;
	int ret;

	/*	Allocate memory	*/
	data = kzalloc(sizeof(struct iqs253_sar_data), GFP_KERNEL);
	input_dev = input_allocate_device();
	if (!data || !input_dev) {
		dev_err(&client->dev, "Failed to allocate memory\n");
		ret = -ENOMEM;
		goto err_free_mem;
	}

	/*	Save the stuctures to be used in the driver	*/
	data->client = client;
	data->input_dev = input_dev;
	data->platform_data = client->dev.platform_data;

	input_dev->name = "Azoteq IQS253 SAR Sensor";
	input_dev->id.bustype = BUS_I2C;
	input_dev->dev.parent = &client->dev;

	/*	Request the RDY line for the IQS253	*/
	gpio_request(RDY_LINE_PIN, "interrupt");
	gpio_direction_input(RDY_LINE_PIN);

	/*	Request Power (PWR - VDDHI) line	*/
	gpio_request(PWR_LINE_PIN, "VDDHI");
	gpio_direction_output(PWR_LINE_PIN, 1);

	/*	Debugging - this GPIO is used for debugging to show a visual
	 *	cue as to which state is active. In practice, antennae would
	 *	be switched ON/OFF where this GPIO is called in this driver
	 */
	gpio_request(53, "led1");
	gpio_direction_output(53, 1);

	gpio_set_value(53, 1);

	/*	Request an interrup on the RDY line	*/
	client->irq = gpio_to_irq(RDY_LINE_PIN);

	/*	Enable the IRQ	*/
	enable_irq(client->irq);

	/* Set driver data */
	input_set_drvdata(input_dev, data);

	/*
	 *	Request the interrupt on a falling trigger
	 *	and only one trigger per falling edge
	 */
	ret = request_threaded_irq(client->irq, NULL, iqs253_sar_interrupt,
			IRQF_TRIGGER_FALLING | IRQF_ONESHOT, DEVICE_NAME, data);

	/*	Register the device */
	ret = input_register_device(data->input_dev);
	if (ret < 0) {
		dev_err(&client->dev, "Failed to register interrupt\n");
		goto err_free_mem;
	}

	/*	Now, assign default values to global variables	*/
	/*	Periodically check for chip reset	*/
	currentState = 0;	/*	Always start at state 0	*/
	setupCounter = 0;	/*	Start initial setup	*/
	doInitialSetup = true;
	initialATI = false;
	eventMode = false;
	touch = false; /*	Assume no touch at first	*/
	reseed = false;
	showReset = false;

	/* Set i2c client data	*/
	i2c_set_clientdata(client, data);

	/*	Setup timer if we miss first setup window	*/
	setup_timer(&stuck, stuck_time, 0);
	mod_timer(&stuck,
		jiffies + msecs_to_jiffies(1000));

	/*	Only swith on VDDHI now	*/
	/*	Switch on VDDHI (Power)	*/
	gpio_set_value(PWR_LINE_PIN, 1);

	return 0;

err_free_mem:
	input_free_device(input_dev);
	kfree(data);
	return ret;
}

static int __devexit iqs253_sar_remove(struct i2c_client *client)
{
	struct iqs253_sar_data *data = i2c_get_clientdata(client);

	free_irq(client->irq, data);

	input_unregister_device(data->input_dev);
	kfree(data);

	i2c_set_clientdata(client, NULL);

	return 0;
}

/*	Standard stucture with the device id for identification	*/
static const struct i2c_device_id iqs253_sar_id[] = {
		{ DEVICE_NAME, 0 },
		{ }
};

MODULE_DEVICE_TABLE(i2c, iqs253_sar_id);

/*
 *	Standard stucture containing the driver
 *	information and procedures
 */
static struct i2c_driver iqs253_sar_driver = {
	.probe = iqs253_sar_probe,
	.remove = __devexit_p(iqs253_sar_remove),
	.driver = {.name = DEVICE_NAME,},
	.id_table = iqs253_sar_id,
};

/**
 *	Gets called from 'board-omap3beagle.c'
 *	when the device I2C bus init is called
 */
static int __init iqs253_sar_init(void)
{
	/*	Add i2c driver to kernel	*/
	printk(KERN_ALERT "Installing IQS253 SAR Sensor Driver");
	return i2c_add_driver(&iqs253_sar_driver);
}

/*	Remove the driver */
static void __exit iqs253_sar_exit(void)
{
	/*	Boolean value used for the initial setup	*/
	setupCounter = 0;	/*	Start initial setup	*/
	doInitialSetup = true;
	initialATI = false;

	/*	Reset the IC - or set the proxsettings again	*/
	gpio_set_value(PWR_LINE_PIN, 0);

	/*	Debug - Reset states	*/
	gpio_set_value(53, 0);

	/*	Delete driver	*/
	i2c_del_driver(&iqs253_sar_driver);
	printk(KERN_ALERT "Delete: - %s\n",
			"IQS253 SAR Sensor driver Deleted! ");
}

module_init(iqs253_sar_init);
module_exit(iqs253_sar_exit);

/* Module information */
MODULE_AUTHOR("Alwino van der Merwe <alwino.vandermerwe@azoteq.com>");
MODULE_DESCRIPTION("SAR Sensor driver for Azoteq IQS253");
MODULE_LICENSE("GPL");
